﻿using System;
using System.Linq;
using System.Security.Claims;
using System.Threading.Tasks;
using Microsoft.AspNetCore.Authentication;
using Microsoft.AspNetCore.Authorization;
using Microsoft.AspNetCore.Mvc;
using Microsoft.IdentityModel.Tokens;
using Tsi1.Api.Infrastructure;
using Tsi1.Api.Models;
using Tsi1.BusinessLayer.Dtos;
using Tsi1.BusinessLayer.Helpers;
using Tsi1.BusinessLayer.Interfaces;

namespace Tsi1.Api.Controllers
{
    [Authorize]
    [Route("api/[controller]")]
    [ApiController]
    public class UserController : ControllerBase
    {
        private readonly IJwtAuthManager _jwtAuthManager;
        private readonly IUserService _userService;
        private readonly IUserTypeService _userTypeService;
        private readonly ITenantService _tenantService;
        private readonly IEmailService _emailService;

        public UserController(
            IJwtAuthManager jwtAuthManager,
            IUserService userService,
            IUserTypeService userTypeService,
            ITenantService tenantService,
            IEmailService emailService)
        {
            _jwtAuthManager = jwtAuthManager;
            _userService = userService;
            _userTypeService = userTypeService;
            _tenantService = tenantService;
            _emailService = emailService;
        }

        [AllowAnonymous]
        [HttpPost("Login")]
        public async Task<IActionResult> Login(LoginRequest request)
        {
            var resultSplit = request.Username.Split("@");

            if (resultSplit.Count() != 2)
            {
                return BadRequest(ErrorMessages.InvalidUsername);
            }

            var userName = resultSplit[0];
            var tenantName = resultSplit[1];
            var tenantId = await _tenantService.GetByName(tenantName);

            if (tenantId.HasError)
            {
                return BadRequest(tenantId.Message);
            }

            var result = await _userService.Authenticate(userName, request.Password, tenantId.Data);
            if (result.HasError)
            {
                return BadRequest(result.Message);
            }

            var user = result.Data;

            var claims = new[]
            {
                new Claim("Id", user.Id.ToString()),
                new Claim("Username", user.Username),
                new Claim("TenantId", user.TenantId.ToString()),
                new Claim(ClaimTypes.Role, user.UserType.Name)
            };

            var jwtResult = _jwtAuthManager.GenerateTokens(user.Username, claims, DateTime.Now);

            return Ok(new LoginResult
            {
                Id = user.Id,
                UserName = user.Username,
                Role = user.UserType.Name,
                AccessToken = jwtResult.AccessToken,
                RefreshToken = jwtResult.RefreshToken.TokenString
            });
        }

        [AllowAnonymous]
        [HttpPost("RefreshToken")]
        public async Task<ActionResult> RefreshToken([FromBody] RefreshTokenRequest request)
        {
            try
            {
                if (string.IsNullOrWhiteSpace(request.RefreshToken))
                {
                    return Unauthorized();
                }

                var accessToken = await HttpContext.GetTokenAsync("Bearer", "access_token");
                var jwtResult = _jwtAuthManager.Refresh(request.RefreshToken, accessToken, DateTime.Now);

                return Ok(new RefreshTokenDto
                {
                    AccessToken = jwtResult.AccessToken,
                    RefreshToken = jwtResult.RefreshToken.TokenString
                });
            }
            catch (SecurityTokenException e)
            {
                return Unauthorized(e.Message); // return 401 so that the client side can redirect the user to login page
            }
        }

        [Authorize(Roles = UserTypes.FacultyAdmin + ", " + UserTypes.UdelarAdmin)]
        [HttpPost("Register")]
        public async Task<IActionResult> Register(UserRegisterDto dto, [FromQuery] int? tenantId = null)
        {
            var myUserType = HttpContext.User.Claims.FirstOrDefault(x => x.Type == ClaimTypes.Role).Value;

            if (myUserType == UserTypes.UdelarAdmin && tenantId == null)
            {
                return BadRequest(string.Format(ErrorMessages.TenantDoesNotExist, tenantId));
            }

            if (myUserType == UserTypes.FacultyAdmin)
            {
                tenantId = int.Parse(HttpContext.User.Claims.FirstOrDefault(x => x.Type == "TenantId").Value);
            }

            var userTypeResult = await _userTypeService.GetById(dto.UserTypeId);

            if (userTypeResult.HasError)
            {
               return BadRequest(userTypeResult.Message);
            }

            var userType = userTypeResult.Data;

            if (myUserType == UserTypes.UdelarAdmin &&
                (userType.Name == UserTypes.Student ||
                userType.Name == UserTypes.Professor))
            {
                return BadRequest(string.Format(ErrorMessages.InvalidUserType, userType.Name));
            }

            if (myUserType == UserTypes.FacultyAdmin &&
                (userType.Name == UserTypes.UdelarAdmin ||
                userType.Name == UserTypes.FacultyAdmin))
            {
                return BadRequest(string.Format(ErrorMessages.InvalidUserType, userType.Name));
            }

            var userServiceResult = await _userService.Create(dto, userType.Name, (int) tenantId);

            if (userServiceResult.HasError)
            {
                BadRequest(userServiceResult.Message);
            }

            return Ok(userServiceResult.Data);
        }

        [Authorize(Roles = UserTypes.FacultyAdmin + ", " + UserTypes.UdelarAdmin)]
        [HttpPut("Modify/{userId}")]
        public async Task<IActionResult> Modify(UserModifyDto dto, int userId)
        {
            var myUserType = HttpContext.User.Claims.FirstOrDefault(x => x.Type == ClaimTypes.Role).Value;

            var userTypeResult = await _userService.GetUserType(userId);

            if (userTypeResult.HasError)
            {
                return BadRequest(userTypeResult.Message);
            }

            if (myUserType == UserTypes.FacultyAdmin)
            {
                var tenantId = int.Parse(HttpContext.User.Claims.FirstOrDefault(x => x.Type == "TenantId").Value);

                var userTenant = await _userService.GetTenant(userId);

                if (userTenant.HasError)
                {
                    return BadRequest(userTenant.Message);
                }

                if (userTenant.Data != tenantId)
                {
                    return BadRequest("No se puede modificar un usuario de otra facultad");
                }
            } 

            var userType = userTypeResult.Data;

            if (myUserType == UserTypes.UdelarAdmin &&
                (userType.Name == UserTypes.Student ||
                userType.Name == UserTypes.Professor))
            {
                return BadRequest(string.Format(ErrorMessages.InvalidUserType, userType.Name));
            }

            if (myUserType == UserTypes.FacultyAdmin &&
                (userType.Name == UserTypes.UdelarAdmin ||
                userType.Name == UserTypes.FacultyAdmin))
            {
                return BadRequest(string.Format(ErrorMessages.InvalidUserType, userType.Name));
            }

            var userServiceResult = await _userService.Modify(dto, userType.Name, userId);

            if (userServiceResult.HasError)
            {
                return BadRequest(userServiceResult.Message);
            }

            if (userServiceResult.Data == false)
            {
                return NotFound(userServiceResult.Message);
            }

            return Ok();
        }

        [Authorize(Roles = UserTypes.FacultyAdmin + ", " + UserTypes.UdelarAdmin)]
        [HttpDelete("Delete/{userId}")]
        public async Task<IActionResult> Delete(int userId)
        {
            var myUserType = HttpContext.User.Claims.FirstOrDefault(x => x.Type == ClaimTypes.Role).Value;

            if (myUserType == UserTypes.FacultyAdmin)
            {
                var tenantId = int.Parse(HttpContext.User.Claims.FirstOrDefault(x => x.Type == "TenantId").Value);

                var userTenant = await _userService.GetTenant(userId);

                if (userTenant.HasError)
                {
                    return BadRequest(userTenant.Message);
                }

                if (userTenant.Data != tenantId)
                {
                    return BadRequest("No se puede borrar un usuario de otra facultad");
                }
            }

            var userTypeResult = await _userService.GetUserType(userId);

            if (userTypeResult.HasError)
            {
                return BadRequest(userTypeResult.Message);
            }

            var userType = userTypeResult.Data;

            if (myUserType == UserTypes.UdelarAdmin &&
                (userType.Name == UserTypes.Student ||
                userType.Name == UserTypes.Professor))
            {
                return BadRequest(string.Format(ErrorMessages.InvalidUserType, userType.Name));
            }

            if (myUserType == UserTypes.FacultyAdmin &&
                (userType.Name == UserTypes.UdelarAdmin ||
                userType.Name == UserTypes.FacultyAdmin))
            {
                return BadRequest(string.Format(ErrorMessages.InvalidUserType, userType.Name));
            }

            var userServiceResult = await _userService.Delete(userId);

            if (userServiceResult.HasError)
            {
                return BadRequest(userServiceResult.Message);
            }

            if (userServiceResult.Data == false)
            {
                return NotFound(userServiceResult.Message);
            }

            return Ok();
        }

        [Authorize(Roles = UserTypes.Student + ", " + UserTypes.Professor + ", " + UserTypes.FacultyAdmin)]
        [HttpGet("GetAll")]
        public async Task<IActionResult> GetAll()
        {
            var tenantId = int.Parse(HttpContext.User.Claims.FirstOrDefault(x => x.Type == "TenantId").Value);

            var result = await _userService.GetAll(tenantId);

            if (result.HasError)
            {
                return BadRequest(result.Message);
            }

            return Ok(result.Data);
        }

        [Authorize(Roles = UserTypes.FacultyAdmin)]
        [HttpGet("GetStudents")]
        public async Task<IActionResult> GetStudents()
        {
            var tenantId = int.Parse(HttpContext.User.Claims.FirstOrDefault(x => x.Type == "TenantId").Value);

            var result = await _userService.GetStudents(tenantId);

            if (result.HasError)
            {
                return BadRequest(result.Message);
            }

            return Ok(result.Data);
        }

        [Authorize(Roles = UserTypes.FacultyAdmin)]
        [HttpGet("GetProfessors")]
        public async Task<IActionResult> GetProfessors()
        {
            var tenantId = int.Parse(HttpContext.User.Claims.FirstOrDefault(x => x.Type == "TenantId").Value);

            var result = await _userService.GetProfessors(tenantId);

            if (result.HasError)
            {
                return BadRequest(result.Message);
            }

            return Ok(result.Data);
        }

        [Authorize(Roles = UserTypes.FacultyAdmin + ", " + UserTypes.UdelarAdmin)]
        [HttpGet("GetAdmins")]
        public async Task<IActionResult> GetAdmins()
        {
            var tenantId = int.Parse(HttpContext.User.Claims.FirstOrDefault(x => x.Type == "TenantId").Value);
            var userType = HttpContext.User.Claims.FirstOrDefault(x => x.Type == ClaimTypes.Role).Value;

            var result = await _userService.GetAdmins(tenantId, userType);

            if (result.HasError)
            {
                return BadRequest(result.Message);
            }

            return Ok(result.Data);
        }

        [Authorize(Roles = UserTypes.Student + ", " + UserTypes.Professor + ", " + UserTypes.FacultyAdmin + ", " + UserTypes.UdelarAdmin)]
        [HttpGet("GetById/{userId}")]
        public async Task<IActionResult> GetById(int userId)
        {
            var result = await _userService.GetById(userId);

            if (result.HasError)
            {
                return BadRequest(result.Message);
            }

            return Ok(result.Data);
        }

        [Authorize(Roles = UserTypes.UdelarAdmin + ", " + UserTypes.FacultyAdmin)]
        [HttpGet("GetUserTypes")]
        public async Task<IActionResult> GetUserTypes()
        {
            var userType = HttpContext.User.Claims.FirstOrDefault(x => x.Type == ClaimTypes.Role).Value;

            var result = await _userTypeService.GetAll(userType);

            if (result.HasError)
            {
                return BadRequest(result.Message);
            }

            return Ok(result.Data);
        }

        [AllowAnonymous]
        [HttpGet("ForgotPassword/{username}")]
        public async Task<IActionResult> ForgotPassword(string username)
        {
            var resultSplit = username.Split("@");

            if (resultSplit.Count() != 2)
            {
                return BadRequest(ErrorMessages.InvalidUsername);
            }

            username = resultSplit[0];
            var tenantName = resultSplit[1];
            var tenantId = await _tenantService.GetByName(tenantName);

            if (tenantId.HasError)
            {
                return BadRequest(tenantId.Message);
            }

            var userResult = await _userService.GetByUsername(username, tenantId.Data);

            if (userResult.HasError)
            {
                return BadRequest(userResult.Message);
            }

            var code = _jwtAuthManager.GenerateVerificationCode(username, DateTime.Now);

            var result = await _emailService.SendVerificationCode(userResult.Data.Email, code);
            if (result.HasError)
            {
                return BadRequest("Ha ocurrido un error");
            }

            return Ok();
        }

        [AllowAnonymous]
        [HttpGet("VerificationCode/{username}/{code}")]
        public async Task<IActionResult> VerificationCode(string username, int code)
        {
            var resultSplit = username.Split("@");

            if (resultSplit.Count() != 2)
            {
                return BadRequest(ErrorMessages.InvalidUsername);
            }

            username = resultSplit[0];
            var tenantName = resultSplit[1];
            var tenantId = await _tenantService.GetByName(tenantName);

            if (tenantId.HasError)
            {
                return BadRequest(tenantId.Message);
            }

            if (!_jwtAuthManager.ValidateVerificationCode(username, code))
            {
                return BadRequest("Código de verificación incorrecto");
            }

            var userResult = await _userService.GetByUsername(username, tenantId.Data);
            if (userResult.HasError)
            {
                return BadRequest(userResult.Message);
            }

            var user = userResult.Data;
            var claims = new[]
            {
                new Claim("Id", user.Id.ToString()),
                new Claim("Username", user.Username),
                new Claim("TenantId", user.TenantId.ToString()),
                new Claim(ClaimTypes.Role, user.UserType.Name)
            };

            var jwtResult = _jwtAuthManager.GenerateTokens(user.Username, claims, DateTime.Now);

            return Ok(new LoginResult
            {
                Id = user.Id,
                UserName = user.Username,
                Role = user.UserType.Name,
                AccessToken = jwtResult.AccessToken,
                RefreshToken = jwtResult.RefreshToken.TokenString
            });
        }

        [HttpPost("RestorePassword")]
        public async Task<IActionResult> RestorePassword(RestorePasswordDto dto)
        {
            var userId = int.Parse(HttpContext.User.Claims.FirstOrDefault(x => x.Type == "Id").Value);

            var result = await _userService.UpdatePassword(userId, dto.Password);
            if (result.HasError)
            {
                return BadRequest(result.Message);
            }

            return Ok();
        }

        [AllowAnonymous]
        [HttpGet("GetAllRefreshTokens")]
        public IActionResult GetAllRefreshTokens()
        {
            return Ok(_jwtAuthManager.GetRefreshTokens());
        }
    }
}
